# -*- coding: utf-8 -*-

from django.test import TestCase

from neomodel import StructuredNode
from chemtrails.neoutils import get_node_for_object, get_node_class_for_model

from tests.utils import flush_nodes
from tests.testapp.autofixtures import BookFixture
from tests.testapp.models import Book


class FlushNodesTestCase(TestCase):
    """
    Make sure the flush_nodes context decorator is working.
    """

    def test_flush_nodes_context_manager(self):

        with flush_nodes():
            book = BookFixture(Book).create_one()

            node = get_node_for_object(book)
            self.assertIsInstance(node, StructuredNode)

            klass = get_node_class_for_model(Book)
            self.assertEqual(len(klass.nodes.all()), 1)

        self.assertEqual(len(klass.nodes.all()), 0)

    def test_flush_nodes_context_decorator(self):

        @flush_nodes()
        def create_nodes():
            book = BookFixture(Book).create_one()

            node = get_node_for_object(book)
            assert isinstance(node, StructuredNode)

            klass = get_node_class_for_model(Book)
            assert len(klass.nodes.all()) == 1

        create_nodes()

        klass = get_node_class_for_model(Book)
        self.assertEqual(len(klass.nodes.all()), 0)




# -*- coding: utf-8 -*-

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db.models import QuerySet
from django.test import TestCase

from chemtrails.contrib.permissions import utils
from chemtrails.contrib.permissions.exceptions import MixedContentTypeError
from chemtrails.contrib.permissions.models import AccessRule
from chemtrails.neoutils import get_node_for_object, get_nodeset_for_queryset
from tests.testapp.autofixtures import Author, AuthorFixture, Book, BookFixture, Store, StoreFixture
from tests.utils import flush_nodes, clear_neo4j_model_nodes

User = get_user_model()


class GetIdentityTestCase(TestCase):
    """
    Testing ``chemtrails.contrib.permissions.utils.get_identity()``.
    """

    def test_get_identity_anonymous_user(self):
        user = AnonymousUser()
        try:
            utils.get_identity(user)
            self.fail('Did not raise NotImplementedError when checking with AnonymousUser.')
        except NotImplementedError as e:
            self.assertEqual(str(e), 'Implement support for AnonymousUser, please!')

    def test_get_identity_user(self):
        user = User.objects.create_user(username='testuser', password='test123.')
        self.assertEqual(utils.get_identity(user), (user, None))

    def test_get_identity_group(self):
        group = Group.objects.create(name='mygroup')
        self.assertEqual(utils.get_identity(group), (None, group))


class GetContentTypeTestCase(TestCase):
    """
    Testing ``chemtrails.contrib.permissions.utils.get_content_type()``.
    """

    def test_get_content_type_from_class(self):
        self.assertEqual(utils.get_content_type(User),
                         ContentType.objects.get_for_model(User))

    def test_get_content_type_from_instance(self):
        user = User.objects.create_user(username='testuser', password='test123.')
        self.assertEqual(utils.get_content_type(user),
                         ContentType.objects.get_for_model(User))


class CheckPermissionsAppLabelTestCase(TestCase):
    """
    Testing ``chemtrails.contrib.permissions.utils.check_permissions_app_label()``.
    """

    def test_check_permissions_app_label_single(self):
        perm = 'testapp.add_book'
        book = BookFixture(Book).create_one()
        self.assertEqual(utils.check_permissions_app_label(perm),
                         (utils.get_content_type(book), {'add_book'}))
        self.assertEqual(utils.check_permissions_app_label(perm),
                         (utils.get_content_type(Book), {'add_book'}))

    def test_check_permissions_app_label_invalid_fails(self):
        perm = 'testapp.invalid_permission'
        self.assertRaisesMessage(
            ContentType.DoesNotExist, 'ContentType matching query does not exist.',
            utils.check_permissions_app_label, permissions=perm)

    def test_check_permissions_app_label_sequence(self):
        perms = ['testapp.add_book', 'testapp.change_book']
        book = BookFixture(Book).create_one()

        ctype, codenames = utils.check_permissions_app_label(perms)
        self.assertEqual(ctype, utils.get_content_type(book))
        self.assertEqual(sorted(codenames), ['add_book', 'change_book'])

    def test_check_permissions_app_label_sequence_fails(self):
        perms = ['testapp.add_book', 'auth.add_user']
        self.assertRaisesMessage(
            MixedContentTypeError, ('Given permissions must have the same app label '
                                    '(testapp != auth).'),
            utils.check_permissions_app_label, permissions=perms)

        perms = ['testapp.add_book', 'testapp.add_store']
        self.assertRaisesMessage(
            MixedContentTypeError, ('Calculated content type from permission "testapp.add_store" '
                                    'store does not match <ContentType: book>.'),
            utils.check_permissions_app_label, permissions=perms)


class GetObjectsForUserTestCase(TestCase):
    """
    Testing ``chemtrails.contrib.permissions.utils.get_objects_for_user()``.
    """

    def setUp(self):
        clear_neo4j_model_nodes()

        BookFixture(Book, generate_m2m={'authors': (2, 2)}).create_one()
        self.user1, self.user2 = User.objects.earliest('pk'), User.objects.latest('pk')
        self.group = Group.objects.create(name='group')

    def tearDown(self):
        clear_neo4j_model_nodes()

    def test_superuser(self):
        self.user1.is_superuser = True
        queryset = Book.objects.all()
        objects = utils.get_objects_for_user(self.user1, ['testapp.change_book'], queryset)
        self.assertEqual(set(queryset), set(objects))

    def test_with_superuser_true(self):
        self.user1.is_superuser = True
        queryset = Book.objects.all()
        objects = utils.get_objects_for_user(self.user1,
                                             ['testapp.change_book'], queryset, with_superuser=True)
        self.assertEqual(set(queryset), set(objects))

    def test_with_superuser_false(self):
        BookFixture(Book, follow_fk=True, generate_m2m={'authors': (1, 1)}).create(count=2)

        user = User.objects.latest('pk')
        user.is_superuser = True

        # `with_superuser=False` requires defined access rules - should yield no results!
        self.assertEqual(set(Book.objects.none()),
                         set(utils.get_objects_for_user(
                             user, ['testapp.change_book'], Book.objects.all(), with_superuser=False)))

        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(user),
                                                ctype_target=utils.get_content_type(Book),
                                                relation_types=[
                                                    'AUTHOR',
                                                    'BOOK'
                                                ])
        perm = Permission.objects.get(content_type__app_label='testapp', codename='change_book')
        access_rule.permissions.add(perm)
        user.user_permissions.add(perm)

        objects = utils.get_objects_for_user(user,
                                             ['testapp.change_book'], Book.objects.all(), with_superuser=False)
        self.assertEqual(set(user.author.book_set.all()), set(objects))

    def test_anonymous(self):
        user = AnonymousUser()
        queryset = Book.objects.all()
        objects = utils.get_objects_for_user(user,
                                             ['testapp.change_book'], queryset)
        self.assertEqual(set(Book.objects.none()), set(objects))

    def test_nonexistent_source_node(self):
        user = User.objects.create_user(username='testuser')

        node = get_node_for_object(user).sync()
        node.delete()

        objects = utils.get_objects_for_user(user, ['testapp.add_book'])
        self.assertEqual(set(objects), set(Book.objects.none()))

    def test_mixed_permissions(self):
        codenames = [
            'testapp.change_book',
            'testapp.change_store'
        ]
        self.assertRaises(MixedContentTypeError, utils.get_objects_for_user, self.user1, codenames)

    def test_mixed_app_label_permissions(self):
        codenames = [
            'testapp.change_book',
            'auth.change_user'
        ]
        self.assertRaises(MixedContentTypeError, utils.get_objects_for_user, self.user1, codenames)

    def test_mixed_ctypes_no_klass(self):
        codenames = [
            'testapp.change_book',
            'auth.change_user'
        ]
        self.assertRaises(MixedContentTypeError, utils.get_objects_for_user, self.user1, codenames)

    def test_mixed_ctypes_with_klass(self):
        codenames = [
            'testapp.change_book',
            'auth.change_user'
        ]
        self.assertRaises(MixedContentTypeError, utils.get_objects_for_user, self.user1, codenames, Book)

    def test_no_app_label_or_klass(self):
        self.assertRaises(ValueError, utils.get_objects_for_user, self.user1, ['change_book'])

    def test_empty_permissions_sequence(self):
        objects = utils.get_objects_for_user(self.user1, [], Book.objects.all())
        self.assertEqual(set(objects), set())

    def test_permissions_single(self):
        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(User),
                                                ctype_target=utils.get_content_type(Group),
                                                relation_types=[
                                                    'GROUPS'
                                                ])
        perm = Permission.objects.get(content_type__app_label='auth', codename='change_group')
        access_rule.permissions.add(perm)
        self.user1.user_permissions.add(perm)

        self.assertEqual(
            set(utils.get_objects_for_user(self.user1, 'auth.change_group')),
            set(utils.get_objects_for_user(self.user1, ['auth.change_group']))
        )

    def test_klass_as_model(self):
        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(User),
                                                ctype_target=utils.get_content_type(Group),
                                                relation_types=[
                                                    'GROUPS'
                                                ])
        perm = Permission.objects.get(content_type__app_label='auth', codename='change_group')
        access_rule.permissions.add(perm)
        self.user1.user_permissions.add(perm)

        self.user1.groups.add(self.group)

        objects = utils.get_objects_for_user(self.user1,
                                             ['auth.change_group'], Group)
        self.assertEqual([obj.name for obj in objects], [self.group.name])

    def test_klass_as_manager(self):
        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(User),
                                                ctype_target=utils.get_content_type(Group),
                                                relation_types=[
                                                    'GROUPS'
                                                ])
        perm = Permission.objects.get(content_type__app_label='auth', codename='change_group')
        access_rule.permissions.add(perm)
        self.user1.user_permissions.add(perm)

        self.user1.groups.add(self.group)

        objects = utils.get_objects_for_user(self.user1,
                                             ['auth.change_group'], Group.objects)
        self.assertEqual([obj.name for obj in objects], [self.group.name])

    def test_klass_as_queryset(self):
        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(User),
                                                ctype_target=utils.get_content_type(Group),
                                                relation_types=[
                                                    'GROUPS'
                                                ])
        perm = Permission.objects.get(content_type__app_label='auth', codename='change_group')
        access_rule.permissions.add(perm)
        self.user1.user_permissions.add(perm)

        self.user1.groups.add(self.group)

        objects = utils.get_objects_for_user(self.user1,
                                             ['auth.change_group'], Group.objects.all())
        self.assertEqual([obj.name for obj in objects], [self.group.name])

    def test_ensure_returns_queryset(self):
        objects = utils.get_objects_for_user(self.user1, ['auth.change_group'])
        self.assertTrue(isinstance(objects, QuerySet))
        self.assertEqual(objects.model, Group)

    def test_single_permission_to_check(self):
        groups = Group.objects.bulk_create([Group(name=name) for name in ['group1', 'group2', 'group3']])
        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(User),
                                                ctype_target=utils.get_content_type(Group),
                                                relation_types=[
                                                    'GROUPS'
                                                ])
        perm = Permission.objects.get(content_type__app_label='auth', codename='change_group')
        access_rule.permissions.add(perm)
        self.user1.user_permissions.add(perm)

        self.user1.groups.add(*groups)

        objects = utils.get_objects_for_user(self.user1, 'auth.change_group')
        self.assertEqual(len(groups), len(objects))
        self.assertEqual(set(groups), set(objects))

    def test_multiple_permissions_to_check(self):
        groups = Group.objects.bulk_create([Group(name=name) for name in ['group1', 'group2', 'group3']])
        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(User),
                                                ctype_target=utils.get_content_type(Group),
                                                relation_types=[
                                                    'GROUPS'
                                                ])
        add_perm = Permission.objects.get(content_type__app_label='auth', codename='add_group')
        change_perm = Permission.objects.get(content_type__app_label='auth', codename='change_group')
        access_rule.permissions.add(*[add_perm, change_perm])
        self.user1.user_permissions.add(*[add_perm, change_perm])

        self.user1.groups.add(*groups)
        objects = utils.get_objects_for_user(self.user1, ['auth.add_group', 'auth.change_group'])

        self.assertEqual(len(groups), len(objects))
        self.assertEqual(set(groups), set(objects))

    def test_multiple_permissions_to_check_requires_staff(self):
        groups = Group.objects.bulk_create([Group(name=name) for name in ['group1', 'group2', 'group3']])
        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(User),
                                                ctype_target=utils.get_content_type(Group),
                                                requires_staff=True,
                                                relation_types=[
                                                    'GROUPS'
                                                ])
        perms = Permission.objects.filter(content_type__app_label='auth', codename__in=['add_group', 'delete_group'])
        access_rule.permissions.add(*perms)

        self.user1.user_permissions.add(*perms)
        self.user1.groups.add(*groups)

        self.user1.is_staff = True
        get_node_for_object(self.user1).sync()  # Sync node in order to write `is_staff` property

        objects = utils.get_objects_for_user(self.user1, ['auth.add_group', 'auth.delete_group'])
        self.assertEqual(set(groups), set(objects))

        self.user2.user_permissions.add(*perms)
        self.user2.groups.add(*groups)

        self.assertFalse(self.user2.is_staff)

        objects = utils.get_objects_for_user(self.user2, ['auth.add_group', 'auth.delete_group'])
        self.assertEqual(set(), set(objects))

    def test_multiple_permissions_to_check_use_groups(self):
        self.group.permissions.add(Permission.objects.get(content_type__app_label='auth', codename='add_group'))
        self.user1.user_permissions.add(Permission.objects.get(content_type__app_label='auth', codename='change_group'))

        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(User),
                                                ctype_target=utils.get_content_type(Group),
                                                relation_types=[
                                                    'GROUPS'
                                                ])
        access_rule.permissions.add(*Permission.objects.filter(content_type__app_label='auth',
                                                               codename__in=['add_group', 'change_group']))

        self.user1.groups.add(self.group)
        objects = utils.get_objects_for_user(self.user1,
                                             ['auth.add_group', 'auth.change_group'], use_groups=True)
        self.assertEqual(set(self.user1.groups.all()), set(objects))

        self.user1.groups.remove(self.group)
        objects = utils.get_objects_for_user(self.user1,
                                             ['auth.add_group', 'auth.change_group'], use_groups=False)
        self.assertEqual(set(), set(objects))

    def test_any_permissions(self):
        groups = Group.objects.bulk_create([Group(name=name) for name in ['group1', 'group2', 'group3']])
        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(User),
                                                ctype_target=utils.get_content_type(Group),
                                                relation_types=[
                                                    'GROUPS'
                                                ])
        perms = Permission.objects.filter(content_type__app_label='auth', codename__in=['add_group', 'change_group'])
        access_rule.permissions.add(*perms)

        self.user1.user_permissions.add(*perms)
        self.user1.groups.add(*groups)

        objects = utils.get_objects_for_user(self.user1, ['auth.add_group', 'auth.delete_group'], any_perm=False)
        self.assertEqual(set(), set(objects))

        objects = utils.get_objects_for_user(self.user1, ['auth.add_group', 'auth.delete_group'], any_perm=True)
        self.assertEqual(set(groups), set(objects))


class GetObjectsForGroupTestCase(TestCase):
    """
    Testing ``chemtrails.contrib.permissions.utils.get_objects_for_group()``.
    """
    pass


class GraphPermissionCheckerTestCase(TestCase):
    """
    Testing ``chemtrails.contrib.permissions.utils.GraphPermissionChecker`` class.
    """

    def test_checker_has_perm_inactive_user(self):
        user = User.objects.create_user(username='testuser', password='test123.', is_active=False)
        checker = utils.GraphPermissionChecker(user)
        self.assertFalse(checker.has_perm(perm=None, obj=None))

    def test_checker_has_perm_is_superuser(self):
        user = User.objects.create_user(username='testuser', password='test123.', is_superuser=True)
        checker = utils.GraphPermissionChecker(user)
        self.assertTrue(checker.has_perm(perm=None, obj=None))

    def test_get_user_filters(self):
        user = User.objects.create_user(username='testuser', password='test123.')
        user.user_permissions.add(*Permission.objects.filter(codename__in=['add_user', 'change_user']))

        # Get user filters for user
        checker = utils.GraphPermissionChecker(user)
        filters = checker.get_user_filters()
        permissions = Permission.objects.filter(**filters)
        self.assertIsInstance(permissions, QuerySet)
        self.assertEqual(permissions.count(), 2)

    def test_get_user_perms(self):
        user = User.objects.create_user(username='testuser', password='test123.')
        user.user_permissions.add(*Permission.objects.filter(codename__in=['add_user', 'change_user']))

        checker = utils.GraphPermissionChecker(user)
        self.assertListEqual(sorted(list(checker.get_user_perms(user))), ['add_user', 'change_user'])

    def test_get_group_filters(self):
        group = Group.objects.create(name='test group')
        group.permissions.add(*Permission.objects.filter(codename__in=['add_user', 'change_user']))

        user = User.objects.create_user(username='testuser', password='test123.')
        user.groups.add(group)

        # Get group filters for group
        checker = utils.GraphPermissionChecker(group)
        filters = checker.get_group_filters()
        permissions = Permission.objects.filter(**filters)
        self.assertIsInstance(permissions, QuerySet)
        self.assertEqual(permissions.count(), 2)

        # Get group filters for use
        checker = utils.GraphPermissionChecker(user)
        filters = checker.get_group_filters()
        permissions = Permission.objects.filter(**filters)
        self.assertIsInstance(permissions, QuerySet)
        self.assertEqual(permissions.count(), 2)

    def test_get_group_perms(self):
        group = Group.objects.create(name='test group')
        group.permissions.add(*Permission.objects.filter(codename__in=['add_user', 'change_user']))

        user = User.objects.create_user(username='testuser', password='test123.')

        checker = utils.GraphPermissionChecker(group)
        self.assertListEqual(sorted(list(checker.get_group_perms(user))), ['add_user', 'change_user'])

    def test_get_perms_user_is_inactive(self):
        user = User.objects.create_user(username='testuser', password='test123.', is_active=False)
        checker = utils.GraphPermissionChecker(user)
        self.assertListEqual(checker.get_perms(user), [])

    def test_get_perms_user_is_superuser(self):
        user = User.objects.create_user(username='testuser', password='test123.', is_superuser=True)
        checker = utils.GraphPermissionChecker(user)
        self.assertListEqual(sorted(checker.get_perms(user)), ['add_user', 'change_user', 'delete_user'])

    def test_get_perms_user_in_group(self):
        group = Group.objects.create(name='test group')
        group.permissions.add(Permission.objects.get(codename='add_user'))

        user = User.objects.create_user(username='testuser', password='test123.')
        user.user_permissions.add(Permission.objects.get(codename='change_user'))
        user.groups.add(group)

        # Make sure we get user and group permissions combined
        checker = utils.GraphPermissionChecker(user)
        self.assertListEqual(sorted(checker.get_perms(user)), ['add_user', 'change_user'])

    def test_get_perms_group(self):
        group = Group.objects.create(name='test group')
        group.permissions.add(Permission.objects.get(codename='add_group'))

        checker = utils.GraphPermissionChecker(group)
        self.assertListEqual(sorted(checker.get_perms(group)), ['add_group'])

    @flush_nodes()
    def test_checker_has_perm_authorized_user(self):
        author = AuthorFixture(Author).create_one()
        user = author.user
        perm = Permission.objects.get(content_type=utils.get_content_type(author), codename='change_author')
        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(user),
                                                ctype_target=utils.get_content_type(author),
                                                relation_types=[
                                                    'AUTHOR'
                                                ])
        user.user_permissions.add(perm)
        access_rule.permissions.add(perm)

        checker = utils.GraphPermissionChecker(user)
        self.assertTrue(checker.has_perm(perm.codename, author))

    @flush_nodes()
    def test_checker_has_perm_authorized_group(self):
        group = Group.objects.create(name='test group')
        user = User.objects.create_user(username='testuser', password='test123.')
        perm = Permission.objects.get(content_type=utils.get_content_type(user), codename='change_user')
        access_rule = AccessRule.objects.create(ctype_source=utils.get_content_type(group),
                                                ctype_target=utils.get_content_type(user),
                                                relation_types=[
                                                    'USER_SET'
                                                ])
        user.groups.add(group)
        group.permissions.add(perm)
        access_rule.permissions.add(perm)

        checker = utils.GraphPermissionChecker(group)
        self.assertTrue(checker.has_perm(perm.codename, user))



# -*- coding: utf-8 -*-
"""
Settings for django-chemtrails are all namespaced in
the CHEMTRAILS setting. For example:
CHEMTRAILS = {
    'ENABLED': True,
    'IGNORE_MODELS: [
        'migrations.migration,
    ]
}
"""
from django.conf import settings

from django.core.signals import setting_changed

# TODO: IGNORE_MODELS should support following formats:
# - app_label: 'migrations.*'
# - model:     '*.migration'
# - specific:  'migrations.migration'
DEFAULTS = {
    'ENABLED': True,
    'MAX_CONNECTION_DEPTH': 1,
    'NAMED_RELATIONSHIPS': True,
    'CONNECT_META_NODES': False,
    'IGNORE_MODELS': [
        'admin.logentry',
        'migrations.migration',
    ],
}


class CSettings:

    def __init__(self, user_settings=None, defaults=None):
        self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError('Invalid setting: \'%s\'' % attr)

        try:
            value = self.user_settings[attr]
        except KeyError:
            value = self.defaults[attr]

        # Cache the setting
        setattr(self, attr, value)
        return value

    @property
    def user_settings(self):
        if not getattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'CHEMTRAILS', {})
        return self._user_settings

chemtrails_settings = CSettings(None, DEFAULTS)


def reload_settings(*args, **kwargs):
    global chemtrails_settings
    setting, value = kwargs['setting'], kwargs['value']
    if setting == 'CHEMTRAILS':
        chemtrails_settings = CSettings(value, DEFAULTS)

setting_changed.connect(reload_settings)



# -*- coding: utf-8 -*-

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import models


class Author(models.Model):
    user = models.OneToOneField('auth.User')
    name = models.CharField(max_length=100)
    age = models.IntegerField()

    def __str__(self):
        return self.name


class Tag(models.Model):
    tag = models.CharField(max_length=20)
    content_type = models.ForeignKey(ContentType, verbose_name='content type',
                                     related_name='content_type_set_for_%(class)s')
    object_pk = models.IntegerField(verbose_name='object ID')
    content_object = GenericForeignKey('content_type', 'object_pk')

    def __str__(self):
        return self.tag


class Publisher(models.Model):
    name = models.CharField(max_length=300)
    num_awards = models.IntegerField()

    class Meta:
        permissions = (
            ('can_publish', 'Can publish'),
        )

    def __str__(self):
        return self.name


class Book(models.Model):
    name = models.CharField(max_length=300)
    pages = models.IntegerField()
    price = models.DecimalField(max_digits=10, decimal_places=2)
    rating = models.FloatField()
    authors = models.ManyToManyField(Author)
    tags = GenericRelation(Tag, object_id_field='object_pk', content_type_field='content_type')
    publisher = models.ForeignKey(Publisher)
    pubdate = models.DateField()

    class Meta:
        permissions = (
            ('view_book', 'Can view book'),
        )

    def __str__(self):
        return self.name


class Store(models.Model):
    name = models.CharField(max_length=300)
    books = models.ManyToManyField(Book)
    bestseller = models.ForeignKey(Book, related_name='bestseller_stores', null=True, blank=True)
    registered_users = models.PositiveIntegerField()

    def __str__(self):
        return self.name



# -*- coding: utf-8 -*-

from django.test import TestCase

from contextlib import ContextDecorator
from neomodel import db


def clear_neo4j_model_nodes():
    db.cypher_query("MATCH (n) WHERE n.type = 'ModelNode' DETACH DELETE n")


class flush_nodes(ContextDecorator):
    """
    Context decorator which will wipe out the all
    ``ModelNode`` nodes on enter and exit.
    """

    def __enter__(self):
        clear_neo4j_model_nodes()

    def __exit__(self, *exc):
        clear_neo4j_model_nodes()


class ChemtrailsTestCase(TestCase):
    """
    Deletes all ``ModelNodes`` from Neo4j in setUp() and tearDown().
    ``MetaNodes`` are created during migration, and are left intact.
    """

    def setUp(self):
        super(ChemtrailsTestCase, self).setUp()
        clear_neo4j_model_nodes()

    def tearDown(self):
        super(ChemtrailsTestCase, self).tearDown()
        clear_neo4j_model_nodes()



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-30 13:06
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0002_tag'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='publisher',
            options={'permissions': (('can_publish', 'Can publish'),)},
        ),
    ]



# -*- coding: utf-8 -*-

from django.contrib.auth.models import Group, Permission
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings
from django.utils import six

from neomodel import *
from neomodel.match import NodeSet

from chemtrails.neoutils import (
    ModelNodeMeta, ModelNodeMixin, MetaNodeMeta, MetaNodeMixin,
    get_meta_node_class_for_model, get_meta_node_for_model,
    get_node_class_for_model, get_node_for_object, get_nodeset_for_queryset
)

from tests.utils import flush_nodes
from tests.testapp.autofixtures import (
    Book, BookFixture, Publisher, PublisherFixture, Store, StoreFixture,
)


class NodeUtilsTestCase(TestCase):
    """
    Test various utility functions for dealing with
    model and node instances.
    """
    def test_get_meta_node_class_for_model(self):
        klass = get_meta_node_class_for_model(Book)
        self.assertTrue(issubclass(klass, StructuredNode))

    @flush_nodes()
    def test_get_meta_node_class_for_concrete_model(self):
        book = BookFixture(Book).create_one()
        klass = get_meta_node_class_for_model(book, for_concrete_model=True)
        self.assertTrue(issubclass(klass, StructuredNode))

    def test_get_node_class_for_model(self):
        klass = get_node_class_for_model(Book)
        self.assertTrue(issubclass(klass, StructuredNode))

    @flush_nodes()
    def test_get_node_class_for_model(self):
        book = BookFixture(Book).create_one()
        klass = get_node_class_for_model(book, for_concrete_model=True)
        self.assertTrue(issubclass(klass, StructuredNode))

    @flush_nodes()
    def test_get_meta_node_for_model(self):
        book_meta = get_meta_node_for_model(Book)
        self.assertIsInstance(book_meta, StructuredNode)

    @flush_nodes()
    def test_get_node_for_object(self):
        store = StoreFixture(Store).create_one(commit=True)
        store_node = get_node_for_object(store)
        self.assertIsInstance(store_node, get_node_class_for_model(Store))

    @flush_nodes()
    def test_get_nodeset_for_queryset(self):
        queryset = Store.objects.filter(pk__in=map(lambda n: n.pk,
                                                   StoreFixture(Store).create(count=2, commit=True)))
        nodeset = get_nodeset_for_queryset(queryset)
        self.assertIsInstance(nodeset, NodeSet)
        for node in nodeset:
            self.assertIsInstance(node, get_node_class_for_model(queryset.model))


class ModelNodeTestCase(TestCase):
    """
    Test that we can create ModelNode instances.
    """
    @flush_nodes()
    def test_create_model_node(self):
        book = BookFixture(Book).create_one()

        @six.add_metaclass(ModelNodeMeta)
        class ModelNode(ModelNodeMixin, StructuredNode):
            class Meta:
                model = Book

        self.assertTrue(issubclass(ModelNode, StructuredNode))
        self.assertIsInstance(ModelNode(instance=book), StructuredNode)

    @flush_nodes()
    def test_create_model_with_content_types(self):
        permission = Permission.objects.latest('pk')

        @six.add_metaclass(ModelNodeMeta)
        class ModelNode(ModelNodeMixin, StructuredNode):
            class Meta:
                model = Permission

        self.assertTrue(issubclass(ModelNode, StructuredNode))
        self.assertIsInstance(ModelNode(instance=permission), StructuredNode)

    def test_create_model_node_declaring_model_in_class(self):

        @six.add_metaclass(ModelNodeMeta)
        class ModelNode(ModelNodeMixin, StructuredNode):
            __metaclass_model__ = Book

            class Meta:
                model = None

        self.assertEqual(ModelNode.Meta.model, Book)

    def test_create_model_node_custom_app_label(self):

        @six.add_metaclass(ModelNodeMeta)
        class ModelNode(ModelNodeMixin, StructuredNode):

            class Meta:
                model = Book
                app_label = 'custom_app_label'

        self.assertEqual(ModelNode.Meta.app_label, 'custom_app_label')

    def test_create_model_node_fails_without_meta_model(self):
        try:
            @six.add_metaclass(ModelNodeMeta)
            class ModelNode(ModelNodeMixin, StructuredNode):

                class Meta:
                    model = None

            self.fail('Did not fail when defining a ModelNode with missing Meta class model.')
        except ValueError as e:
            self.assertEqual(str(e), '%s.Meta.model attribute cannot be None.' % 'ModelNode')

    def test_create_model_node_fails_without_meta_class(self):
        try:
            @six.add_metaclass(ModelNodeMeta)
            class ModelNode(ModelNodeMixin, StructuredNode):
                pass

            self.fail('Did not fail when defining a ModelNode without a Meta class.')
        except ImproperlyConfigured as e:
            self.assertEqual(str(e), '%s must implement a Meta class.' % 'ModelNode')

    def test_save_existing_node_is_updated(self):
        group = Group.objects.create(name='a group')

        node1 = get_node_for_object(group)
        self.assertEqual(group.name, node1.name)

        group.name = 'still the same group'
        group.save()

        node2 = get_node_for_object(group)
        self.assertEqual(group.name, node2.name)

        self.assertEqual(node1.id, node2.id)

    def test_sync_existing_node_is_updated(self):
        group = Group.objects.create(name='a group')

        node = get_node_for_object(group)
        self.assertEqual(group.name, node.name)

        group.name = 'another name'
        group.save()

        node.sync()
        self.assertEqual(group.name, node.name)


class MetaNodeTestCase(TestCase):
    """
    Test that we can create "singleton" meta nodes
    """
    def test_create_meta_node(self):

        @six.add_metaclass(MetaNodeMeta)
        class MetaNode(MetaNodeMixin, StructuredNode):

            class Meta:
                model = Book

        self.assertTrue(issubclass(MetaNode, StructuredNode))
        self.assertIsInstance(MetaNode(), StructuredNode)

    def test_create_meta_node_declaring_model_in_class(self):

        @six.add_metaclass(MetaNodeMeta)
        class MetaNode(MetaNodeMixin, StructuredNode):
            __metaclass_model__ = Book

            class Meta:
                model = None

        self.assertEqual(MetaNode.Meta.model, Book)

    def test_create_meta_node_custom_app_label(self):

        @six.add_metaclass(MetaNodeMeta)
        class MetaNode(MetaNodeMixin, StructuredNode):

            class Meta:
                model = Book
                app_label = 'custom_app_label'

        self.assertEqual(MetaNode.Meta.app_label, 'custom_app_label')

    def test_create_meta_node_custom_permissions(self):

        @six.add_metaclass(MetaNodeMeta)
        class MetaNode(MetaNodeMixin, StructuredNode):
            class Meta:
                model = Publisher

        self.assertEqual(MetaNode.Meta.model, Publisher)

    def test_create_meta_node_fails_without_meta_model(self):
        try:
            @six.add_metaclass(MetaNodeMeta)
            class MetaNode(MetaNodeMixin, StructuredNode):

                class Meta:
                    model = None

            self.fail('Did not fail when defining a MetaNode with missing Meta class model.')
        except ValueError as e:
            self.assertEqual(str(e), '%s.Meta.model attribute cannot be None.' % 'MetaNode')

    def test_create_meta_node_fails_without_meta_class(self):
        try:
            @six.add_metaclass(MetaNodeMeta)
            class MetaNode(MetaNodeMixin, StructuredNode):
                pass

            self.fail('Did not fail when defining a MetaNode without a Meta class.')
        except ImproperlyConfigured as e:
            self.assertEqual(str(e), '%s must implement a Meta class.' % 'MetaNode')

    @flush_nodes()
    @override_settings(CHEMTRAILS={
        'CONNECT_META_NODES': True
    })
    def test_connected_meta_node(self):
        book = BookFixture(Book).create_one()
        meta = get_meta_node_for_model(Book).sync()

        # FIXME: Settings object is not updated when using override_settings


class GraphMapperTestCase(TestCase):
    """
    Test that node relationships are mapped correctly
    """

    def test_get_property_class_for_field(self):
        from django.contrib.contenttypes.fields import GenericRelation
        from django.contrib.postgres.fields import (
            ArrayField, HStoreField, JSONField,
            IntegerRangeField, BigIntegerRangeField, FloatRangeField,
            DateTimeRangeField, DateRangeField
        )
        from django.db import models

        class CustomField(object):
            pass

        klass = get_node_class_for_model(Book)

        self.assertEqual(klass.get_property_class_for_field(models.ForeignKey), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.ForeignKey), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.OneToOneField), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.ManyToManyField), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.ManyToOneRel), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.OneToOneRel), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(models.ManyToManyRel), RelationshipTo)
        self.assertEqual(klass.get_property_class_for_field(GenericRelation), RelationshipTo)

        self.assertEqual(klass.get_property_class_for_field(models.AutoField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.BigAutoField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.BooleanField), BooleanProperty)
        self.assertEqual(klass.get_property_class_for_field(models.CharField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.CommaSeparatedIntegerField), ArrayProperty)
        self.assertEqual(klass.get_property_class_for_field(models.DateField), DateProperty)
        self.assertEqual(klass.get_property_class_for_field(models.DateTimeField), DateTimeProperty)
        self.assertEqual(klass.get_property_class_for_field(models.DecimalField), FloatProperty)
        self.assertEqual(klass.get_property_class_for_field(models.DurationField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.EmailField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.FilePathField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.FileField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.FloatField), FloatProperty)
        self.assertEqual(klass.get_property_class_for_field(models.GenericIPAddressField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.IntegerField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.IPAddressField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.NullBooleanField), BooleanProperty)
        self.assertEqual(klass.get_property_class_for_field(models.PositiveIntegerField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.PositiveSmallIntegerField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.SlugField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.SmallIntegerField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.TextField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.TimeField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(models.URLField), StringProperty)
        self.assertEqual(klass.get_property_class_for_field(models.UUIDField), StringProperty)

        # Test special fields
        self.assertEqual(klass.get_property_class_for_field(ArrayField), ArrayProperty)
        self.assertEqual(klass.get_property_class_for_field(HStoreField), JSONProperty)
        self.assertEqual(klass.get_property_class_for_field(JSONField), JSONProperty)

        # Test undefined fields by inspecting their base classes.
        self.assertEqual(klass.get_property_class_for_field(models.BigIntegerField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(IntegerRangeField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(BigIntegerRangeField), IntegerProperty)
        self.assertEqual(klass.get_property_class_for_field(FloatRangeField), FloatProperty)
        self.assertEqual(klass.get_property_class_for_field(DateTimeRangeField), DateTimeProperty)
        self.assertEqual(klass.get_property_class_for_field(DateRangeField), DateProperty)

        # Test unsupported field
        self.assertRaisesMessage(
            NotImplementedError, 'Unsupported field. Field CustomField is currently not supported.',
            klass.get_property_class_for_field, CustomField)

    @flush_nodes()
    def test_sync_related_branch(self):
        queryset = Store.objects.filter(pk__in=map(lambda n: n.pk,
                                                   StoreFixture(Store).create(count=1, commit=True)))
        store_nodeset = get_nodeset_for_queryset(queryset, sync=True, max_depth=1)
        for store in store_nodeset:
            store_obj = store.get_object()

            if store_obj.bestseller:
                self.assertEqual(store.bestseller.get(), get_node_for_object(store_obj.bestseller))

            self.assertEqual(len(store.books.all()), store_obj.books.count())
            for book in store.books.all():
                book_obj = book.get_object()
                self.assertTrue(store in book.store_set.all())
                self.assertEqual(book.publisher.get(), get_node_for_object(book_obj.publisher))
                self.assertEqual(len(book.store_set.all()), book_obj.store_set.count())
                self.assertEqual(len(book.bestseller_stores.all()), book_obj.bestseller_stores.count())
                self.assertEqual(len(book.authors.all()), book_obj.authors.count())

                for author in book.authors.all():
                    author_obj = author.get_object()
                    self.assertTrue(book in author.book_set.all())

                    user = author.user.get()
                    self.assertEqual(user, get_node_for_object(author_obj.user).sync())
                    self.assertEqual(author, user.author.get())




# -*- coding: utf-8 -*-

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.test import TestCase, modify_settings

from rest_framework import status
from rest_framework.permissions import DjangoObjectPermissions
from rest_framework.serializers import ModelSerializer
from rest_framework.test import APIRequestFactory, force_authenticate
from rest_framework.viewsets import ModelViewSet

from chemtrails.contrib.permissions.models import AccessRule
from chemtrails.contrib.permissions.rest_framework.filters import ChemoPermissionsFilter
from chemtrails.contrib.permissions.utils import get_content_type

User = get_user_model()
factory = APIRequestFactory()


class ChemoPermissionsFilterTestCase(TestCase):

    backend = 'chemtrails.contrib.permissions.backends.ChemoPermissionsBackend'

    def setUp(self):
        Group.objects.bulk_create([Group(name=name) for name in ['group1', 'group2', 'group3']])

        class GroupSerializer(ModelSerializer):
            class Meta:
                model = Group
                fields = '__all__'

        class GroupViewSet(ModelViewSet):
            queryset = Group.objects.all()
            serializer_class = GroupSerializer
            permission_classes = [DjangoObjectPermissions]
            filter_backends = [ChemoPermissionsFilter]

        self.user = User.objects.create_user(username='testuser', password='test123.')
        self.perm = Permission.objects.create(content_type=get_content_type(Group),
                                              name='Can view group', codename='view_group')
        self.access_rule = AccessRule.objects.create(ctype_source=get_content_type(User),
                                                     ctype_target=get_content_type(Group),
                                                     is_active=True,
                                                     relation_types=[
                                                         'GROUPS'
                                                     ])
        self.view = GroupViewSet

        self.patched_settings = modify_settings(
            AUTHENTICATION_BACKENDS={'append': self.backend}
        )
        self.patched_settings.enable()

    def tearDown(self):
        self.patched_settings.disable()

    def test_filter_get_list(self):
        groups = Group.objects.filter(name__in=['group1', 'group2'])
        self.user.user_permissions.add(self.perm)
        self.user.groups.add(*groups)
        self.access_rule.permissions.add(self.perm)

        request = factory.get(path='', format='json')
        force_authenticate(request, self.user)

        # User should now be able to see group1 and group2, but not group3
        response = self.view.as_view(actions={'get': 'list'})(request)
        self.assertEqual(len(response.data), len(groups))
        for result in response.data:
            self.assertTrue(result['name'] in groups.values_list('name', flat=True))

    def test_filter_get_detail(self):
        group1, group2 = (Group.objects.get(name='group1'),
                          Group.objects.get(name='group2'))
        self.user.user_permissions.add(self.perm)
        self.user.groups.add(group1)
        self.access_rule.permissions.add(self.perm)

        request = factory.get(path='', format='json')
        force_authenticate(request, self.user)

        response = self.view.as_view(actions={'get': 'retrieve'})(request, pk=group1.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['name'], group1.name)

        response = self.view.as_view(actions={'get': 'retrieve'})(request, pk=group2.pk)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.data, {'detail': 'Not found.'})

    def test_filter_post(self):
        self.user.user_permissions.add(
            Permission.objects.get(content_type__app_label='auth', codename='add_group'))

        request = factory.post(path='', data={'name': 'new group'}, format='json')
        force_authenticate(request, self.user)

        response = self.view.as_view(actions={'post': 'create'})(request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data['name'], 'new group')

    def test_filter_put(self):
        group = Group.objects.get(name='group1')
        perms = Permission.objects.filter(content_type__app_label='auth', codename__in=['view_group', 'change_group'])
        self.user.groups.add(group)
        self.user.user_permissions.add(*perms)
        self.access_rule.permissions.add(*perms)

        request = factory.put(path='', data={'name': 'put group'}, format='json')
        force_authenticate(request, self.user)

        response = self.view.as_view(actions={'put': 'update'})(request, pk=group.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['name'], 'put group')

    def test_filter_patch(self):
        group = Group.objects.get(name='group1')
        perms = Permission.objects.filter(content_type__app_label='auth', codename__in=['view_group', 'change_group'])
        self.user.groups.add(group)
        self.user.user_permissions.add(*perms)
        self.access_rule.permissions.add(*perms)

        request = factory.patch(path='', data={'name': 'patch group'}, format='json')
        force_authenticate(request, self.user)

        response = self.view.as_view(actions={'patch': 'update'})(request, pk=group.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['name'], 'patch group')

    def test_filter_delete(self):
        group = Group.objects.create(name='some group')
        perms = Permission.objects.filter(content_type__app_label='auth', codename__in=['view_group', 'delete_group'])
        self.user.groups.add(group)
        self.user.user_permissions.add(*perms)
        self.access_rule.permissions.add(*perms)

        request = factory.delete(path='', format='json')
        force_authenticate(request, self.user)

        response = self.view.as_view(actions={'delete': 'destroy'})(request, pk=group.pk)
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)



# -*- coding: utf-8 -*-
from django.conf.urls import url, include
from rest_framework import routers
from .views import BookViewSet

router = routers.DefaultRouter()
router.register(r'books', BookViewSet)

urlpatterns = [
    url(r'^api-auth/', include('rest_framework.urls', namespace='rest_framework')),
    url(r'^', include(router.urls))
]



# -*- coding: utf-8 -*-

import random
from datetime import timedelta, date

from django.contrib.auth import get_user_model

from autofixture import generators, register, AutoFixture
from tests.testapp.models import Author, Publisher, Book, Store, Tag


class PublisherGenerator(generators.Generator):
    publishers = [
        'Addison-Wesley', 'Aladdin Paperbacks', 'Anvil Press Poetry',
        'Basic Books', 'Blackstaff Press', 'Blake Publishing', 'Butterworth-Heinemann',
        'Cambridge University Press', 'Century', 'Crocker & Brewster',
        'Dedalus Books', 'Dick and Fitzgerald', 'Dorchester Publishing',
        'Fairview Press', 'Four Walls Eight Windows'
    ]

    def generate(self):
        return random.choice(self.publishers)


class UserGenerator(generators.Generator):

    USER_MODEL = get_user_model()

    def generate(self):
        defaults = {
            self.USER_MODEL.USERNAME_FIELD: '{name}{num1}{num2}'.format(
                name=generators.LastNameGenerator().generate(),
                num1=generators.SmallIntegerGenerator().generate(),
                num2=generators.SmallIntegerGenerator().generate()
            ).lower(),
            'password': 'test123.'
        }
        return self.USER_MODEL.objects.create_user(**defaults)


class AuthorFixture(AutoFixture):
    field_values = {
        'user': UserGenerator(),
        'name': generators.FirstNameGenerator(),
        'age': generators.PositiveIntegerGenerator(min_value=20, max_value=80)
    }
register(Author, AuthorFixture, overwrite=True)


class PublisherFixture(AutoFixture):
    field_values = {
        'name': PublisherGenerator(),
        'num_awards': generators.PositiveIntegerGenerator(min_value=0, max_value=10)
    }
register(Publisher, PublisherFixture, overwrite=True)


class BookFixture(AutoFixture):
    generate_fk = True
    generate_m2m = {'authors': (1, 2)}
    field_values = {
        'name': generators.StringGenerator(min_length=5, max_length=15),
        'pages': generators.PositiveIntegerGenerator(min_value=200, max_value=2000),
        'price': generators.PositiveDecimalGenerator(max_digits=10, decimal_places=2),
        'rating': generators.FloatGenerator(max_value=10, min_value=0, decimal_digits=2),
        'pubdate': generators.DateGenerator(min_date=date.today() - timedelta(days=36500),
                                            max_date=date.today())
    }
    # FIXME: This doesn't work!
    # def post_process_instance(self, instance, commit):
    #     values = ('sci-fi', 'drama', 'fantasy', 'romance', 'self help', 'satire')
    #     Tag(content_object=instance, tag=random.choice(values)).save()
    #     return instance
register(Book, BookFixture, overwrite=True)


class StoreFixture(AutoFixture):
    generate_fk = True
    generate_m2m = {'books': (2, 4)}
    field_values = {
        'name': generators.StringGenerator(min_length=5, max_length=15),
        'registered_users': generators.PositiveSmallIntegerGenerator(10, 1000)
    }

    def post_process_instance(self, instance, commit):
        # Make sure the bestseller is in the store!
        if instance.bestseller:
            instance.books.add(instance.bestseller)
        return instance
register(Store, StoreFixture, overwrite=True)


# class TagFixture(AutoFixture):
#     from django.contrib.contenttypes.models import ContentType
#     field_values = {
#         'tag': generators.ChoicesGenerator(values=('sci-fi', 'drama', 'fantasy', 'romance', 'self help', 'satire')),
#         'content_type': generators.StaticGenerator(value=ContentType.objects.get_for_model(Book)),
#         'object_pk': generators.ChoicesGenerator(values=Book.objects.values_list('pk', flat=True)
#                                                  or list(BookFixture(Book).create_one().pk))
#     }
# register(Tag, TagFixture, overwrite=True)



# -*- coding: utf-8 -*-

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from rest_framework import serializers
from chemtrails.contrib.permissions.serializers import (
    AccessRuleSerializer, ContentTypeIdentityField, PermissionIdentityField
)


class AccessRuleSerializerTestCase(TestCase):

    def test_access_rule_serializer_valid_data(self):
        serializer = AccessRuleSerializer(data={
            'ctype_source': 'auth.user',
            'ctype_target': 'testapp.book',
            'permissions': [
                'testapp.add_book',
                'testapp.change_book'
            ]
        })
        self.assertTrue(serializer.is_valid())
        self.assertEqual(serializer.errors, {})

        self.assertEqual(serializer.validated_data['ctype_source'],
                         ContentType.objects.get(app_label='auth', model='user'))
        self.assertEqual(serializer.validated_data['ctype_target'],
                         ContentType.objects.get(app_label='testapp', model='book'))

        for perm in Permission.objects.filter(codename__in=['add_book', 'change_book']):
            self.assertTrue(perm in serializer.validated_data['permissions'])


class ContentTypeIdentityFieldTestCase(TestCase):

    def setUp(self):

        class TestSerializer(serializers.Serializer):
            ctype = ContentTypeIdentityField(queryset=ContentType.objects.all())

        self.Serializer = TestSerializer
        self.ctype = ContentType.objects.get(app_label='auth', model='user')

    def test_serialize_valid_input(self):
        serializer = self.Serializer(data={'ctype': 'auth.user'})
        self.assertTrue(serializer.is_valid())
        self.assertEqual(serializer.validated_data['ctype'], self.ctype)

    def test_serialize_invalid_data_type(self):
        serializer = self.Serializer(data={'ctype': 1})
        self.assertFalse(serializer.is_valid())
        self.assertTrue('ctype' in serializer.errors)

        try:
            serializer.is_valid(raise_exception=True)
            self.fail('Did not raise ValidationError when serializing invalid data type using `raise_exception=True`')
        except serializers.ValidationError as e:
            self.assertEqual(
                str(e), "{'ctype': ['Incorrect type. Expected content type string identifier, received int.']}")

    def test_serialize_invalid_length_content_type_string(self):
        serializer = self.Serializer(data={'ctype': 'auth'})
        self.assertFalse(serializer.is_valid())
        self.assertTrue('ctype' in serializer.errors)

        try:
            serializer.is_valid(raise_exception=True)
            self.fail('Did not raise ValidationError when serializing invalid content '
                      'type string using `raise_exception=True`')
        except serializers.ValidationError as e:
            self.assertEqual(str(e), "{'ctype': ['Incorrect length. Expected content type string, "
                                     "separated by punctuation. Received \"auth\".']}")

    def test_serializer_non_existent_content_type(self):
        serializer = self.Serializer(data={'ctype': 'non.existent'})
        self.assertFalse(serializer.is_valid())
        self.assertTrue('ctype' in serializer.errors)

        try:
            serializer.is_valid(raise_exception=True)
            self.fail('Did not raise exception when serializing non-existent content '
                      'type string using using `raise_exception=True`')
        except serializers.ValidationError as e:
            self.assertEqual(str(e), "{'ctype': ['Invalid content type \"non.existent\" - object does not exist.']}")


class PermissionIdentityFieldTestCase(TestCase):

    def setUp(self):

        class TestSerializer(serializers.Serializer):
            perm = PermissionIdentityField(queryset=Permission.objects.all())

        self.Serializer = TestSerializer
        self.ctype = ContentType.objects.get(app_label='auth', model='user')
        self.permission = Permission.objects.get(content_type=self.ctype, codename='add_user')
        self.perm_string = '{app_label}.{codename}'.format(app_label=self.ctype.app_label,
                                                           codename=self.permission.codename)

    def test_serialize_valid_permission_input(self):
        serializer = self.Serializer(data={'perm': self.perm_string})
        self.assertTrue(serializer.is_valid())
        self.assertEqual(serializer.validated_data['perm'], self.permission)

    def test_serialize_invalid_permission_data_type(self):
        serializer = self.Serializer(data={'perm': 1})
        self.assertFalse(serializer.is_valid())
        self.assertTrue('perm' in serializer.errors)

        try:
            serializer.is_valid(raise_exception=True)
            self.fail('Did not raise ValidationError when serializing invalid data type using `raise_exception=True`')
        except serializers.ValidationError as e:
            self.assertEqual(
                str(e), "{'perm': ['Incorrect type. Expected permission string identifier, received int.']}")

    def test_serialize_invalid_length_permission_string(self):
        serializer = self.Serializer(data={'perm': 'add_user'})
        self.assertFalse(serializer.is_valid())
        self.assertTrue('perm' in serializer.errors)

        try:
            serializer.is_valid(raise_exception=True)
            self.fail('Did not raise ValidationError when serializing invalid permission '
                      'string using `raise_exception=True`')
        except serializers.ValidationError as e:
            self.assertEqual(str(e), "{'perm': ['Incorrect length. Expected permission string identifier, "
                                     "separated by punctuation. Received \"add_user\".']}")

    def test_serializer_non_existent_permission(self):
        serializer = self.Serializer(data={'perm': 'auth.can_levitate'})
        self.assertFalse(serializer.is_valid())
        self.assertTrue('perm' in serializer.errors)

        try:
            serializer.is_valid(raise_exception=True)
            self.fail('Did not raise exception when serializing non-existent permission '
                      'string using using `raise_exception=True`')
        except serializers.ValidationError as e:
            self.assertEqual(str(e), "{'perm': ['Invalid permission \"auth.can_levitate\" "
                                     "- object does not exist.']}")



# -*- coding: utf-8 -*-

from rest_framework.viewsets import ModelViewSet
from rest_framework.serializers import ModelSerializer
from rest_framework.permissions import DjangoObjectPermissions

from chemtrails.contrib.permissions.rest_framework.filters import ChemoPermissionsFilter

from tests.testapp.models import Book


class BookSerializer(ModelSerializer):
    class Meta:
        model = Book
        fields = '__all__'


class BookViewSet(ModelViewSet):
    queryset = Book.objects.all()
    serializer_class = BookSerializer
    permission_classes = [DjangoObjectPermissions]
    filter_backends = [ChemoPermissionsFilter]




# -*- coding: utf-8 -*-

import json

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType

from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITestCase, APIClient

from chemtrails.contrib.permissions.models import AccessRule


class GraphWidgetAPIViews(APITestCase):

    def setUp(self):
        super(GraphWidgetAPIViews, self).setUp()
        self.client = APIClient(enforce_csrf_checks=True)

    def test_get_access_rule_list(self):
        response = self.client.get(reverse('admin:accessrule-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_access_rule_detail(self):
        access_rule = AccessRule.objects.create(
            ctype_source=ContentType.objects.get_by_natural_key('auth', 'user'),
            ctype_target=ContentType.objects.get_by_natural_key('testapp', 'book'),
            is_active=True,
            relation_types=['BOOK']
        )
        access_rule.permissions.add(*list(Permission.objects.filter(codename__in=['add_book', 'change_book'])))
        response = self.client.get(reverse('admin:accessrule-detail', kwargs={'pk': access_rule.pk}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_post_access_rule(self):
        response = self.client.post(reverse('admin:accessrule-list'), data={
            'ctype_source': 'auth.user',
            'ctype_target': 'testapp.book',
            'permissions': [
                'testapp.add_book',
                'testapp.change_book',
                'testapp.delete_book'
            ],
            'relation_types': ['BOOK']
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_patch_access_rule(self):
        access_rule = AccessRule.objects.create(
            ctype_source=ContentType.objects.get_by_natural_key('auth', 'user'),
            ctype_target=ContentType.objects.get_by_natural_key('testapp', 'book'),
            is_active=True
        )
        response = self.client.patch(reverse('admin:accessrule-detail', kwargs={'pk': access_rule.pk}),
                                     data={'is_active': False}, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        content = json.loads(response.content.decode())
        self.assertFalse(content['is_active'])



# -*- coding: utf-8 -*-

from django.db.models.signals import post_save, pre_delete, m2m_changed
from django.test import TestCase

from chemtrails.neoutils import get_node_class_for_model
from chemtrails.signals.handlers import post_save_handler, pre_delete_handler, m2m_changed_handler

from tests.testapp.autofixtures import Author, Book, BookFixture
from tests.utils import flush_nodes


class PostSaveHandlerTestCase(TestCase):

    @flush_nodes()
    def test_create_new_object_is_synced(self):
        post_save.disconnect(post_save_handler, dispatch_uid='chemtrails.signals.handlers.post_save_handler')
        post_save.connect(post_save_handler, dispatch_uid='post_save_handler.test')
        try:
            book = BookFixture(Book).create_one()
            klass = get_node_class_for_model(Book)

            self.assertEqual(book.pk, klass.nodes.get(pk=book.pk).pk)
            self.assertEqual(1, len(klass.nodes.has(authors=True, publisher=True)))
        finally:
            post_save.connect(post_save_handler, dispatch_uid='chemtrails.signals.handlers.post_save_handler')
            post_save.disconnect(post_save_handler, dispatch_uid='post_save_handler.test')

    @flush_nodes()
    def test_delete_object_is_deleted(self):
        pre_delete.disconnect(pre_delete_handler, dispatch_uid='chemtrails.signals.handlers.pre_delete_handler')
        pre_delete.connect(pre_delete_handler, dispatch_uid='pre_delete_handler.test')
        try:
            book = BookFixture(Book).create_one()
            klass = get_node_class_for_model(Book)
            pk = book.pk
            try:
                book.delete()
                klass.nodes.get(pk=pk)
                self.fail('Did not raise when trying to get non-existent book node.')
            except klass.DoesNotExist as e:
                self.assertEqual(str(e), "{'pk': %d}" % pk)
        finally:
            pre_delete.connect(pre_delete_handler, dispatch_uid='chemtrails.signals.handlers.pre_delete_handler')
            pre_delete.disconnect(pre_delete_handler, dispatch_uid='pre_delete_handler.test')

    @flush_nodes()
    def test_m2m_changed_post_add_handler(self):
        m2m_changed.disconnect(m2m_changed_handler, dispatch_uid='chemtrails.signals.handlers.m2m_changed_handler')
        m2m_changed.connect(m2m_changed_handler, dispatch_uid='m2m_changed_handler.test')
        try:
            pass
        finally:
            m2m_changed.connect(m2m_changed_handler, dispatch_uid='chemtrails.signals.handlers.m2m_changed_handler')
            m2m_changed.disconnect(m2m_changed_handler, dispatch_uid='m2m_changed_handler.test')

    @flush_nodes()
    def test_m2m_changed_pre_clear(self):
        m2m_changed.disconnect(m2m_changed_handler, dispatch_uid='chemtrails.signals.handlers.m2m_changed_handler')
        m2m_changed.connect(m2m_changed_handler, dispatch_uid='m2m_changed_handler.test')
        try:
            book = BookFixture(Book).create_one()
            self.assertTrue(len(get_node_class_for_model(Book).nodes.has(authors=True)) >= 1)

            book.authors.clear()
            self.assertEqual(0, len(get_node_class_for_model(Book).nodes.has(authors=True)))
            self.assertEqual(0, len(get_node_class_for_model(Author).nodes.has(book_set=True)))
        finally:
            m2m_changed.connect(m2m_changed_handler, dispatch_uid='chemtrails.signals.handlers.m2m_changed_handler')
            m2m_changed.disconnect(m2m_changed_handler, dispatch_uid='m2m_changed_handler.test')



# -*- coding: utf-8 -*-
from django.core.exceptions import ValidationError
from django.test import TestCase

from chemtrails.neoutils import query


class QueryFunctionsTestCase(TestCase):
    """
    Test the various query functions in chemtrails.neoutils.query.
    """

    def test_get_relationship_types(self):
        expected = ['ACCESSRULE', 'ACCESSRULE_CTYPE_SOURCE_SET', 'ACCESSRULE_CTYPE_TARGET_SET', 'ACCESSRULE_PERMISSIONS',
                    'AUTHOR', 'AUTHORS', 'BESTSELLER', 'BESTSELLER_STORES', 'BOOK', 'BOOKS', 'CONTENT_TYPE',
                    'CONTENT_TYPE_SET_FOR_TAG', 'CTYPE_SOURCE', 'CTYPE_TARGET', 'GROUP', 'GROUPS', 'LOGENTRY',
                    'PERMISSION', 'PERMISSIONS', 'PUBLISHER', 'STORE', 'TAGS', 'USER', 'USER_PERMISSIONS', 'USER_SET']
        self.assertEqual(expected, query.get_relationship_types())

    def test_get_node_relationship_types(self):
        result = query.get_node_relationship_types()
        self.assertIsInstance(result, dict)
        for key, val in result.items():
            self.assertIsInstance(key, str)
            self.assertIsInstance(val, list)

    def test_get_node_relationship_types_with_params(self):
        params = {'type': 'MetaNode', 'app_label': 'auth'}
        result = query.get_node_relationship_types(params)
        self.assertIsInstance(result, dict)
        self.assertTrue(len(result), 6)

    def test_get_node_permissions(self):
        result = query.get_node_permissions()
        self.assertIsInstance(result, list)

    def test_validate_cypher_statement(self):
        is_valid, _ = query.validate_cypher('MATCH (n) RETURN n')
        self.assertTrue(is_valid)

        is_valid, _ = query.validate_cypher('MATCH RETURN n')
        self.assertFalse(is_valid)

        try:
            query.validate_cypher('Invalid cypher statement', raise_exception=True)
            self.fail('Did not raise ValidationError when validating invalid statement.')
        except ValidationError as e:
            self.assertEqual(str(e), "['Failed to validate Cypher statement']")



# -*- coding: utf-8 -*-

from django.test import TestCase, override_settings
from chemtrails import settings


class SettingsTestCase(TestCase):

    def test_default_settings(self):
        self.assertEqual(settings.ENABLED, True)
        self.assertEqual(settings.MAX_CONNECTION_DEPTH, 1)
        self.assertEqual(settings.NAMED_RELATIONSHIPS, True)
        self.assertEqual(settings.CONNECT_META_NODES, False)
        self.assertEqual(settings.IGNORE_MODELS, ['admin.logentry', 'migrations.migration'])

    @override_settings(CHEMTRAILS={
        'ENABLED': False,
        'NAMED_RELATIONSHIPS': False,
        'IGNORE_MODELS': [
            'auth.user'
        ]
    })
    def test_override_settings(self):
        from chemtrails.settings import chemtrails_settings as settings  # Need to re-import after override
        self.assertEqual(settings.ENABLED, False)
        self.assertEqual(settings.NAMED_RELATIONSHIPS, False)
        self.assertEqual(settings.CONNECT_META_NODES, False)
        self.assertEqual(settings.IGNORE_MODELS, ['auth.user'])

    def test_getting_invalid_setting(self):
        try:
            settings.INVALID_SETTING
            self.fail('Did not raise AttributeError when declaring an invalid setting')
        except AttributeError as e:
            self.assertEqual(str(e), 'Invalid setting: \'INVALID_SETTING\'')

    def test_setting_changed_signal_updates_global_settings_object(self):
        try:
            raise NotImplementedError
        except NotImplementedError as e:
            self.assertIsInstance(e, NotImplementedError)



# -*- coding: utf-8 -*-

from django.core.management.base import BaseCommand
from django.utils.translation import ungettext

from chemtrails.neoutils import get_nodeset_for_queryset

from tests.testapp.autofixtures import StoreFixture
from tests.testapp.models import Store


class Command(BaseCommand):
    help = 'Creates a bookstore test graph'

    def add_arguments(self, parser):
        parser.add_argument('count', type=int, help='Number of separate graph trees to create.')

        parser.add_argument(
            '--connection-depth', '-c',
            dest='depth',
            default=1,
            type=int,
            help='Maximum depth of recursive connections to be made.'
        )

    def handle(self, *args, **options):
        self.stdout.write(self.style.NOTICE('Hang tight, this might take a little while...'))
        count = options['count']
        depth = options['depth']
        queryset = Store.objects.filter(pk__in=map(lambda n: n.pk,
                                                   StoreFixture(Store).create(count=count, commit=True)))
        get_nodeset_for_queryset(queryset, sync=True, max_depth=depth)
        self.stdout.write(self.style.SUCCESS(ungettext(
            'Successfully created %(count)d bookstore graph.\nCheck it out in the Neo4j web console!',
            'Successfully created %(count)d bookstore graphs.\nCheck them out in the Neo4j web console!', count
        ) % {'count': count}))



# -*- coding: utf-8 -*-

from chemtrails.settings import chemtrails_settings as settings

default_app_config = 'chemtrails.apps.ChemtrailsConfig'

__title__ = 'django-chemtrails'
__version__ = '0.0.2'
__author__ = 'Rolf Haavard Blindheim'
__license__ = 'MIT License'

VERSION = __version__

__all__ = [
    'settings'
]



# -*- coding: utf-8 -*-

import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = 'not-so-secret'
DEBUG = True

ALLOWED_HOSTS = ['*']
CORS_ORIGIN_ALLOW_ALL = True

LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True
LOGIN_REDIRECT_URL = '/api/'  # Redirect to the browseable API

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    # Dependencies
    'corsheaders',     # Required for chemtrails.contrib.permissions API views.
    'rest_framework',  # Required for easing interactive development.

    # For testing
    'autofixture',
    'tests.testapp',

    'chemtrails',
    'chemtrails.contrib.permissions'
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

ROOT_URLCONF = 'tests.urls'
WSGI_APPLICATION = 'tests.wsgi.application'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': os.environ.get('DB_NAME', 'postgres'),
        'USER': os.environ.get('DB_USER', 'postgres'),
        'PASSWORD': os.environ.get('DB_PASSWORD', 'password'),
        'HOST': os.environ.get('DB_HOST', 'localhost'),
        'PORT': os.environ.get('DB_PORT', 5432)
    }
}

AUTH_PASSWORD_VALIDATORS = [
    {'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator'},
    {'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator'},
    {'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator'},
    {'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator'},
]
AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend',
    # 'chemtrails.contrib.permissions.backends.Neo4jPermissionBackend'
)

STATIC_URL = '/static/'

# Logging
# os.environ['NEOMODEL_CYPHER_DEBUG'] = 'Yes, please!'  # Uncomment to enable cypher debug output
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'default': {
            'format': '%(levelname)s %(asctime)s %(module)s %(message)s'
        }
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'default'
        }
    },
    'loggers': {
        'chemtrails': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': True,
        },
        'neomodel': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': True,
        },
    }
}

try:
    from .local_settings import *
except ImportError:
    pass



# -*- coding: utf-8 -*-

import os

from django.apps import AppConfig
from django.conf import settings
from django.db.models.signals import m2m_changed, post_migrate, post_save, pre_delete

from neomodel import config

config.AUTO_INSTALL_LABELS = False


class ChemtrailsConfig(AppConfig):
    name = 'chemtrails'

    def ready(self):
        from .signals.handlers import (
            m2m_changed_handler, post_migrate_handler,
            post_save_handler, pre_delete_handler
        )

        m2m_changed.connect(receiver=m2m_changed_handler,
                            dispatch_uid='chemtrails.signals.handlers.m2m_changed_handler')
        post_save.connect(receiver=post_save_handler,
                          dispatch_uid='chemtrails.signals.handlers.post_save_handler')
        pre_delete.connect(receiver=pre_delete_handler,
                           dispatch_uid='chemtrails.signals.handlers.pre_delete_handler')

        post_migrate.connect(receiver=post_migrate_handler,
                             dispatch_uid='neomodel.core.post_migrate_handler')

        # Neo4j config
        config.DATABASE_URL = getattr(settings, 'NEO4J_BOLT_URL',
                                      os.environ.get('NEO4J_BOLT_URL', config.DATABASE_URL))
        config.FORCE_TIMEZONE = getattr(settings, 'NEO4J_FORCE_TIMEZONE',
                                        os.environ.get('NEO4J_FORCE_TIMEZONE', False))



# -*- coding: utf-8 -*-

from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from chemtrails.contrib.permissions.models import (
    AccessRule,
    get_node_relations_choices, get_node_permissions_choices
)


class ChoicesHelperFunctionsTestCase(TestCase):
    """
    Test various functions for getting choices based on Neo4j data.
    """

    def test_get_node_relations_choices(self):
        choices = get_node_relations_choices()

        self.assertIsInstance(choices, list)
        for item in choices:
            self.assertIsInstance(item, tuple)
            self.assertEqual(len(item), 2)

    def test_get_node_permissions_choices(self):
        choices = get_node_permissions_choices()

        self.assertIsInstance(choices, list)
        for item in choices:
            self.assertIsInstance(item, tuple)
            self.assertEqual(len(item), 2)



# -*- coding: utf-8 -*-

from django.test import TestCase


class ArrayChoiceFieldTestCase(TestCase):

    def test_display(self):
        pass

    def test_to_python(self):
        pass

    def test_validate(self):
        pass



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-04-06 21:59
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0003_publisher_custom_permission'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='book',
            options={'permissions': (('view_book', 'Can view book'),)},
        ),
    ]



# -*- coding: utf-8 -*-

from collections import Sequence


def get_model_string(model):
    """
    :param model: model
    :returns: <app_label>.<model_name> string representation for the model
    """
    return "{app_label}.{model_name}".format(app_label=model._meta.app_label, model_name=model._meta.model_name)


def flatten(sequence):
    """
    Flatten an arbitrary nested sequence.
    Example usage:
      >> my_list = list(flatten(nested_lists))
    :param sequence: A nested list or tuple.
    :returns: A generator with all values in a flat structure.
    """
    for i in sequence:
        if isinstance(i, Sequence) and not isinstance(i, (str, bytes)):
            yield from flatten(i)
        else:
            yield i



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-10 12:10
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Author',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('age', models.IntegerField()),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Book',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=300)),
                ('pages', models.IntegerField()),
                ('price', models.DecimalField(decimal_places=2, max_digits=10)),
                ('rating', models.FloatField()),
                ('pubdate', models.DateField()),
                ('authors', models.ManyToManyField(to='testapp.Author')),
            ],
        ),
        migrations.CreateModel(
            name='Publisher',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=300)),
                ('num_awards', models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='Store',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=300)),
                ('registered_users', models.PositiveIntegerField()),
                ('bestseller', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='bestseller_stores', to='testapp.Book')),
                ('books', models.ManyToManyField(to='testapp.Book')),
            ],
        ),
        migrations.AddField(
            model_name='book',
            name='publisher',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='testapp.Publisher'),
        ),
    ]



from django.conf.urls import url, include
from django.contrib import admin

import autofixture
autofixture.autodiscover()

urlpatterns = [
    url(r'^admin/', admin.site.urls),
    url(r'^api/', include('tests.testapp.urls'))
]



# -*- coding: utf-8 -*-

from .settings import INSTALLED_APPS, MIDDLEWARE

INSTALLED_APPS += [
    'django_extensions',
    'debug_toolbar',
]

MIDDLEWARE = [
    'debug_toolbar.middleware.DebugToolbarMiddleware'
] + MIDDLEWARE

DEBUG_TOOLBAR_PANELS = [
    'debug_toolbar.panels.versions.VersionsPanel',
    'debug_toolbar.panels.timer.TimerPanel',
    'debug_toolbar.panels.settings.SettingsPanel',
    'debug_toolbar.panels.headers.HeadersPanel',
    'debug_toolbar.panels.request.RequestPanel',
    'debug_toolbar.panels.sql.SQLPanel',
    'debug_toolbar.panels.staticfiles.StaticFilesPanel',
    'debug_toolbar.panels.templates.TemplatesPanel',
    'debug_toolbar.panels.cache.CachePanel',
    'debug_toolbar.panels.signals.SignalsPanel',
    'debug_toolbar.panels.logging.LoggingPanel',
    'debug_toolbar.panels.redirects.RedirectsPanel',
]

DEBUG_TOOLBAR_CONFIG = {
    'SHOW_COLLAPSED': True
}

CHEMTRAILS = {
    'MAX_CONNECTION_DEPTH': 1
}



# -*- coding: utf-8 -*-

import re
import os

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)

setup(
    name='django-chemtrails',
    version=get_version('chemtrails'),
    description='Graphing Django in Neo4j',
    long_description='Graphing Django in Neo4j slightly longer.',
    author='Rolf Håvard Blindheim',
    author_email='rhblind@gmail.com',
    url='https://github.com/inonit/django-chemtrails',
    download_url='https://github.com/inonit/django-chemtrails.git',
    license='MIT License',
    packages=[
        'chemtrails',
    ],
    include_package_data=True,
    install_requires=[
        'Django>=1.8.0',
        'django-cors-headers',
        'libcypher-parser-python>=0.0.2',
        'neomodel>=3.1.0,<3.2.0',
        'psycopg2>=2.6.2',
        'requests[security]',
    ],
    tests_require=[
        'nose',
        'coverage',
        'django-autofixture',
        'djangorestframework'
    ],
    zip_safe=False,
    test_suite='tests.run_tests.start',
    classifiers=[
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import nose
import logging

logging.disable(logging.INFO)  # Disable debug logging when running the test suite.


def start(argv=None):
    sys.exitfunc = lambda: sys.stderr.write('Shutting down...\n')

    if argv is None:
        argv = [
            'nosetests',
            '--verbose',
            '--with-coverage',
            '--cover-html', '--cover-html-dir=.htmlcov',
            '--cover-erase',
            '--cover-branches',
            '--cover-package=chemtrails',
        ]

    nose.run_exit(argv=argv, defaultTest=os.path.abspath(os.path.dirname(__file__)))

if __name__ == '__main__':
    start(sys.argv)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-23 14:23
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('testapp', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('tag', models.CharField(max_length=20)),
                ('object_pk', models.IntegerField(verbose_name='object ID')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='content_type_set_for_tag', to='contenttypes.ContentType', verbose_name='content type')),
            ],
        ),
    ]



# -*- coding: utf-8 -*-

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission, AnonymousUser, Group
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase, modify_settings

from chemtrails.contrib.permissions.backends import ChemoPermissionsBackend
from chemtrails.contrib.permissions.models import AccessRule
from chemtrails.neoutils import get_nodeset_for_queryset

from tests.utils import clear_neo4j_model_nodes
from tests.testapp.autofixtures import Author, AuthorFixture, Store, StoreFixture

User = get_user_model()


class ChemoPermissionsBackendTestCase(TestCase):

    backend = 'chemtrails.contrib.permissions.backends.ChemoPermissionsBackend'

    def setUp(self):
        super(ChemoPermissionsBackendTestCase, self).setUp()
        self.patched_settings = modify_settings(
            AUTHENTICATION_BACKENDS={'append': self.backend}
        )
        self.patched_settings.enable()
        clear_neo4j_model_nodes()

    def tearDown(self):
        super(ChemoPermissionsBackendTestCase, self).tearDown()
        self.patched_settings.disable()
        clear_neo4j_model_nodes()

    def test_authenticate(self):
        kwargs = {
            'username': 'testuser',
            'password': 'test123.'
        }
        User.objects.create_user(**kwargs)

        backend = ChemoPermissionsBackend()
        self.assertIsNone(backend.authenticate(**kwargs))

    def test_get_all_permissions_user(self):
        permissions = Permission.objects.filter(content_type__app_label='auth',
                                                codename__in=[
                                                    'add_user', 'change_user',
                                                    'add_group', 'change_group'
                                                ])
        user = User.objects.create_user(username='testuser', password='test123.')
        group = Group.objects.create(name='group')
        user.user_permissions.add(*permissions)

        backend = ChemoPermissionsBackend()

        self.assertEqual(
            set(permissions.filter(content_type__model='group').values_list('codename', flat=True)),
            set(backend.get_all_permissions(user, group))
        )
        self.assertEqual(
            set(permissions.filter(content_type__model='user').values_list('codename', flat=True)),
            set(backend.get_all_permissions(user, user))
        )

    def test_get_all_permissions_group(self):
        permissions = Permission.objects.filter(content_type__app_label='auth',
                                                codename__in=[
                                                    'add_user', 'change_user',
                                                    'add_group', 'change_group'
                                                ])
        user = User.objects.create_user(username='testuser', password='test123.')
        group = Group.objects.create(name='group')
        group.user_set.add(user)
        group.permissions.add(*permissions)

        backend = ChemoPermissionsBackend()

        self.assertEqual(
            set(permissions.filter(content_type__model='group').values_list('codename', flat=True)),
            set(backend.get_all_permissions(user, group))
        )
        self.assertEqual(
            set(permissions.filter(content_type__model='user').values_list('codename', flat=True)),
            set(backend.get_all_permissions(user, user))
        )

    def test_get_all_permissions_mixed_user_and_group(self):
        permissions = Permission.objects.filter(content_type__app_label='auth',
                                                codename__in=[
                                                    'add_user', 'change_user',
                                                    'add_group', 'change_group'
                                                ])
        user = User.objects.create_user(username='testuser', password='test123.')
        group = Group.objects.create(name='group')
        group.user_set.add(user)

        user.user_permissions.add(*permissions.filter(content_type__model='user'))
        group.permissions.add(*permissions.filter(content_type__model='group'))

        backend = ChemoPermissionsBackend()

        self.assertEqual(
            set(permissions.filter(content_type__model='group').values_list('codename', flat=True)),
            set(backend.get_all_permissions(user, group))
        )
        self.assertEqual(
            set(permissions.filter(content_type__model='user').values_list('codename', flat=True)),
            set(backend.get_all_permissions(user, user))
        )

    def test_get_all_permissions_non_model_instance(self):
        class FooBar:
            pass

        obj = FooBar()
        user = User.objects.create_user(username='testuser', password='test123.')
        backend = ChemoPermissionsBackend()
        self.assertEqual(set(), backend.get_all_permissions(user, obj))

    def test_get_all_permissions_anonymous_user(self):
        class FooBar:
            pass

        obj = FooBar()
        user = AnonymousUser()
        backend = ChemoPermissionsBackend()
        self.assertEqual(set(), backend.get_all_permissions(user, obj))

    def test_has_perm_non_model_instance(self):
        class FooBar:
            pass

        obj = FooBar()
        user = User.objects.create_user(username='testuser', password='test123.')
        backend = ChemoPermissionsBackend()
        self.assertFalse(backend.has_perm(user, 'app_label.codename', obj))

    def test_has_perm_anonymous_user(self):
        user = AnonymousUser()
        group = Group.objects.create(name='group')

        backend = ChemoPermissionsBackend()
        self.assertFalse(backend.has_perm(user, 'app_label.codename', group))

    def test_check_permission(self):
        user = User.objects.create_user(username='testuser', password='test123.')

        group = Group.objects.create(name='group')
        group.user_set.add(user)

        access_rule = AccessRule.objects.create(
            ctype_source=ContentType.objects.get_for_model(User),
            ctype_target=ContentType.objects.get_for_model(Group),
            relation_types=['GROUPS']
        )
        perm = Permission.objects.get(content_type__app_label='auth', codename='add_group')
        group.permissions.add(perm)
        access_rule.permissions.add(perm)

        backend = ChemoPermissionsBackend()
        self.assertTrue(backend.has_perm(user, 'add_group', group))       # Permission codename only
        self.assertTrue(backend.has_perm(user, 'auth.add_group', group))  # Full permission string

    def test_user_has_perm_single_relation_type(self):
        author = AuthorFixture(Author).create_one()
        permission = Permission.objects.get(codename='change_author')
        author.user.user_permissions.add(permission)
        access_rule = AccessRule.objects.create(
            ctype_source=ContentType.objects.get_for_model(author.user),
            ctype_target=ContentType.objects.get_for_model(author),
            relation_types=['AUTHOR']
        )
        access_rule.permissions.add(permission)
        self.assertTrue(author.user.has_perm('testapp.change_author', author))

    def test_user_has_perm_multiple_relation_types(self):
        # Set up a bookstore graph
        store = StoreFixture(Store).create_one()
        get_nodeset_for_queryset(Store.objects.filter(pk=store.pk), sync=True)
        permissions = Permission.objects.filter(codename__in=['add_store', 'change_store', 'delete_store'])

        user = get_user_model().objects.latest('pk')
        user.user_permissions.add(*list(permissions))

        # Create an access rule which allows 'add_store' and 'change_store',
        # but not 'delete_store'
        access_rule = AccessRule.objects.create(
            ctype_source=ContentType.objects.get_for_model(user),
            ctype_target=ContentType.objects.get_for_model(Store),
            relation_types=['AUTHOR', 'BOOK', 'STORE']
        )
        access_rule.permissions.add(*list(permissions.exclude(codename='delete_store')))

        self.assertTrue(user.has_perm('testapp.add_store', store))
        self.assertTrue(user.has_perm('testapp.change_store', store))
        self.assertFalse(user.has_perm('testapp.delete_store', store))



# -*- coding: utf-8 -*-

from django.test import TestCase

from neomodel import db

from chemtrails.neoutils import get_node_for_object
from chemtrails.neoutils.managers import PathManager

from tests.testapp.autofixtures import get_user_model, Book, BookFixture
from tests.utils import flush_nodes


class PathManagerTestCase(TestCase):

    def setUp(self):
        self.node = get_node_for_object(BookFixture(Book).create_one())

    @flush_nodes()
    def tearDown(self):
        pass

    def test_node_has_path_manager(self):
        self.assertIsInstance(self.node.paths, PathManager)

    def test_path_manager_build_path_query(self):
        user = get_user_model().objects.latest('pk')
        query = self.node.paths.add('AUTHORS').add('USER', **{
            'pk': user.pk,
            'username': user.username,
            'is_active': user.is_active
        }).get_path()
        result, meta = db.cypher_query(query)
        self.assertTrue(len(result))
        self.assertEqual(meta, ('path',))

    def test_path_manager_build_match_query(self):
        query = self.node.paths.add('AUTHORS').add('USER').get_match()
        result, meta = db.cypher_query(query)
        self.assertTrue(len(result))



# -*- coding: utf-8 -*-

from django.contrib import admin
from tests.testapp.models import *


@admin.register(Author)
class AuthorAdmin(admin.ModelAdmin):
    pass


@admin.register(Publisher)
class PublisherAdmin(admin.ModelAdmin):
    pass


@admin.register(Book)
class BookAdmin(admin.ModelAdmin):
    pass


@admin.register(Store)
class StoreAdmin(admin.ModelAdmin):
    pass
